# SEC_Node Properties

MANDATORY = "mandatory"
OPTIONAL = "optional"

# mandatory
MODULES = "modules"
EQUIPMENT_ID = "equipment_id"
DESCRIPTION = "description"

# optional
FIRMWARE = "firmware"
IMPLEMENTOR = "implementor"
TIMEOUT = "timeout"


# Module Properties

# mandatory
ACCESSIBLES = "accessibles"
# DESCRIPTION = "description"
INTERFACE_CLASSES = "interface_classes"

# optional
VISIBILITY = "visibility"
GROUP = "group"
MEANING = "meaning"
# IMPLEMENTOR = "implementor"
IMPLEMENTATION = "implementation"
FEATURES = "features"


# Accessible Properties

# mandatory
# DESCRIPTION = "description"
READONLY = "readonly"
DATAINFO = "datainfo"

# optional
# GROUP = "group"
# VISIBILITY = "visibility"
CONSTANT = "constant"


# Property Dicionaries

NODE_PROPERTIES = {
    MANDATORY: [MODULES, EQUIPMENT_ID, DESCRIPTION],
    OPTIONAL: [FIRMWARE, IMPLEMENTOR, TIMEOUT],
}

MODULE_PROPERTIES = {
    MANDATORY: [ACCESSIBLES, DESCRIPTION, INTERFACE_CLASSES],
    OPTIONAL: [VISIBILITY, GROUP, MEANING, IMPLEMENTOR, IMPLEMENTATION, FEATURES],
}

ACESSIBLE_PROPERTIES = {
    MANDATORY: [DESCRIPTION, READONLY, DATAINFO],
    OPTIONAL: [GROUP, VISIBILITY, CONSTANT],
}



# mypy: disable-error-code="attr-defined"
import asyncio

from bluesky.protocols import Triggerable
from frappy.errors import ImpossibleError
from ophyd_async.core import SignalR, SignalX

from secop_ophyd.SECoPDevices import (
    SECoPCMDDevice,
    SECoPMoveableDevice,
    SECoPNodeDevice,
)


async def test_stop_cmd(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):
    cryo: SECoPMoveableDevice = cryo_node_internal_loop.cryo

    stat = cryo.set(15)

    await asyncio.sleep(3)

    await cryo.stop(success=True)

    await stat

    assert cryo._stopped is True

    await cryo_node_internal_loop.disconnect_async()


async def test_stop_no_sucess_cmd(cryo_sim, cryo_node_internal_loop: SECoPNodeDevice):
    cryo: SECoPMoveableDevice = cryo_node_internal_loop.cryo

    stat = cryo.set(15)

    rt_error = False

    await asyncio.sleep(3)

    try:
        await cryo.stop(False)
        await stat

    except RuntimeError:
        rt_error = True

    assert cryo._stopped is True
    assert rt_error is True

    await cryo_node_internal_loop.disconnect_async()


async def test_struct_inp_cmd(nested_struct_sim, nested_node: SECoPNodeDevice):
    test_cmd: SECoPCMDDevice = nested_node.ophy_struct.test_cmd_CMD

    input_dict = {"name": "test_name", "id": 900, "sort": False}

    await test_cmd.argument.set(input_dict)  # type: ignore

    res: SignalR = test_cmd.result  # type: ignore

    run_obj: SignalX = test_cmd.commandx

    stat = run_obj.trigger()

    await stat

    reading_res = await res.read()
    print(reading_res)
    assert isinstance(reading_res.get(res.name)["value"], int)

    await nested_node.disconnect_async()


def test_triggerable(nested_struct_sim, nested_node: SECoPNodeDevice):
    test_cmd: SECoPCMDDevice = nested_node.ophy_struct.test_cmd_CMD

    assert isinstance(test_cmd, Triggerable)


async def test_secop_error_on_cmd(nested_struct_sim, nested_node: SECoPNodeDevice):
    test_cmd: SECoPCMDDevice = nested_node.ophy_struct.test_cmd_CMD

    error_triggered = False
    # Triggers SECoP Error
    input_dict = {"name": "bad_name", "id": 900, "sort": False}

    await test_cmd.argument.set(input_dict)  # type: ignore

    res: SignalR = test_cmd.result  # type: ignore

    run_obj: SignalX = test_cmd.commandx

    try:
        stat = run_obj.trigger()
        await stat

    except ImpossibleError:
        error_triggered = True

    assert error_triggered is True

    reading_res = await res.read()
    assert reading_res.get(res.name)["value"] is None

    await nested_node.disconnect_async()


async def test_secop_triggering_cmd_dev(
    nested_struct_sim, nested_node: SECoPNodeDevice
):
    test_cmd: SECoPCMDDevice = nested_node.ophy_struct.test_cmd_CMD

    input_dict = {"name": "test_name", "id": 900, "sort": False}

    await test_cmd.argument.set(input_dict)  # type: ignore

    res: SignalR = test_cmd.result  # type: ignore

    stat = test_cmd.trigger()

    await stat

    reading_res = await res.read()
    assert isinstance(reading_res.get(res.name)["value"], int)

    await nested_node.disconnect_async()



import inspect
import re
from importlib import import_module, reload
from inspect import Signature
from pathlib import Path


class Method:
    def __init__(self, cmd_name: str, description: str, cmd_sign: Signature) -> None:
        self.sig_str: str

        raw_sig_str: str = str(cmd_sign)

        raw_sig_str = raw_sig_str.replace("typing.", "")

        if "self" in raw_sig_str:
            self.sig_str = raw_sig_str
        else:
            self.sig_str = "(self, " + raw_sig_str[1:]
        self.name: str = cmd_name
        self.description: str = description

    def __str__(self) -> str:
        code = ""
        code += "    @abstractmethod \n"
        code += f"    def {self.name}{self.sig_str}:\n"
        code += f'      """{self.description}"""'

        return code


class GenNodeCode:
    """Generates A Python Class for a given SECoP_Node_Device instance. This allows
    autocompletiion and type hinting in IDEs, this is needed since the attributes of
    the generated Ophyd devices are only known at runtime.
    """

    ModName: str = "genNodeClass"
    node_mod = None
    module_folder_path: Path | None = None

    def __init__(self, path: str | None = None, log=None):
        """Instantiates GenNodeCode, internally all atrribues on a node and module level
        are collected. Additionally all the needed imports are collected in a dict
        """
        # prevent circular import
        from secop_ophyd.SECoPDevices import (
            SECoPBaseDevice,
            SECoPCommunicatorDevice,
            SECoPMoveableDevice,
            SECoPNodeDevice,
            SECoPReadableDevice,
            SECoPTriggerableDevice,
            SECoPWritableDevice,
        )

        self.dimport: dict[str, set[str]] = {}
        self.dmodule: dict = {}
        self.dnode: dict = {}

        self.import_string: str = ""
        self.mod_cls_string: str = ""
        self.node_cls_string: str = ""

        self.log = log

        if path is not None:
            self.module_folder_path = Path(path)

        # resulting Class is supposed to be abstract
        self.add_import("abc", "ABC")
        self.add_import("abc", "abstractmethod")
        self.add_import("typing", "Any")

        mod_path = self.ModName

        if self.module_folder_path is not None:
            str_path = str(self.module_folder_path)
            rep_slash = str_path.replace("/", ".").replace("\\", ".")
            mod_path = f"{rep_slash}.{self.ModName}"

        try:
            self.node_mod = import_module(mod_path)
        except ModuleNotFoundError:
            if self.log is None:
                print("no code generated yet, building from scratch")
            else:
                self.log.info("no code generated yet, building from scratch")
            return

        modules = inspect.getmembers(self.node_mod)

        results = filter(lambda m: inspect.isclass(m[1]), modules)
        for class_symbol, class_obj in results:
            module = class_obj.__module__

            if module == self.ModName:
                # Node Classes

                def get_attrs(source: str) -> list[tuple[str, str]]:
                    source_list: list[str] = source.split("\n")
                    # remove first line
                    source_list.pop(0)
                    source_list.pop()

                    # remove whitespace
                    source_list = [attr.replace(" ", "") for attr in source_list]
                    # split at colon
                    attrs: list[tuple[str, str]] = []
                    for attr in source_list:
                        parts = attr.split(":", maxsplit=1)
                        if len(parts) == 2:
                            attrs.append((parts[0], parts[1]))

                    return attrs

                if issubclass(class_obj, SECoPNodeDevice):
                    attrs = get_attrs(inspect.getsource(class_obj))

                    bases = [base.__name__ for base in class_obj.__bases__]

                    self.add_node_class(class_symbol, bases, attrs)
                    continue

                if issubclass(
                    class_obj,
                    (
                        SECoPBaseDevice,
                        SECoPCommunicatorDevice,
                        SECoPMoveableDevice,
                        SECoPReadableDevice,
                        SECoPWritableDevice,
                        SECoPTriggerableDevice,
                    ),
                ):

                    attributes = []

                    for attr_name, attr in class_obj.__annotations__.items():
                        attributes.append((attr_name, attr.__name__))

                    methods = []

                    for method_name, method in class_obj.__dict__.items():
                        if callable(method) and not method_name.startswith("__"):

                            method_source = inspect.getsource(method)

                            match = re.search(
                                r"\s*def\s+\w+\s*\(.*\).*:\s*", method_source
                            )
                            if match:
                                function_body = method_source[match.end() :]
                                description_list = function_body.split('"""', 2)
                                description = description_list[1]
                            else:
                                raise Exception(
                                    "could not extract description function body"
                                )

                            methods.append(
                                Method(
                                    method_name, description, inspect.signature(method)
                                )
                            )

                    bases = [base.__name__ for base in class_obj.__bases__]

                    self.add_mod_class(class_symbol, bases, attributes, methods)
                    continue

            else:
                self.add_import(module, class_symbol)

    def add_import(self, module: str, class_str: str):
        """adds an Import to the import dict

        :param module: Python Module of the dict
        :type module: str
        :param class_str: Class that is to be imported
        :type class_str: str
        """
        if self.dimport.get(module):
            self.dimport[module].add(class_str)
        else:
            self.dimport[module] = {class_str}

    def add_mod_class(
        self,
        module_cls: str,
        bases: list[str],
        attrs: list[tuple[str, str]],
        cmd_plans: list[Method],
    ):
        """adds module class to the module dict

        :param module_cls: name of the new Module class
        :type module_cls: str
        :param bases: bases the new class is derived from
        :type bases: list[str]
        :param attrs: list of attributes of the class
        :type attrs: tuple[str, str]
        """
        self.dmodule[module_cls] = {"bases": bases, "attrs": attrs, "plans": cmd_plans}

    def add_node_class(
        self, node_cls: str, bases: list[str], attrs: list[tuple[str, str]]
    ):
        self.dnode[node_cls] = {"bases": bases, "attrs": attrs}

    def _write_imports_string(self):
        self.import_string = ""

        # Collect imports required for type hints (Node)
        for mod, cls_set in self.dimport.items():
            if len(cls_set) == 1 and list(cls_set)[0] == "":
                self.import_string += f"import {mod} \n"
                continue

            cls_string = ", ".join(cls_set)

            self.import_string += f"from {mod} import {cls_string} \n"
        self.import_string += "\n\n"

    def _write_mod_cls_string(self):
        self.mod_cls_string = ""

        for mod_cls, cls_dict in self.dmodule.items():
            # Generate the Python code for each Module class
            bases = ", ".join(cls_dict["bases"])

            self.mod_cls_string += f"class {mod_cls}({bases}):\n"
            # write Attributes
            for attr_name, attr_type in cls_dict["attrs"]:
                self.mod_cls_string += f"    {attr_name}: {attr_type}\n"
            self.mod_cls_string += "\n"
            # write abstract methods
            for plan in cls_dict["plans"]:
                self.mod_cls_string += str(plan)

            self.mod_cls_string += "\n\n"

    def _write_node_cls_string(self):
        self.node_cls_string = ""

        for node_cls, cls_dict in self.dnode.items():
            # Generate the Python code for each Module class
            bases = ", ".join(cls_dict["bases"])

            self.node_cls_string += f"class {node_cls}({bases}):\n"
            for attr_name, attr_type in cls_dict["attrs"]:
                self.node_cls_string += f"    {attr_name}: {attr_type}\n"

            self.node_cls_string += "\n\n"

    def write_gen_node_class_file(self):
        self._write_imports_string()
        self._write_mod_cls_string()
        self._write_node_cls_string()

        code = ""

        code += self.import_string
        code += self.mod_cls_string
        code += self.node_cls_string

        # Write the generated code to a .py file

        if self.module_folder_path is None:
            filep = Path(f"{self.ModName}.py")
        else:
            filep = self.module_folder_path / f"{self.ModName}.py"
        with open(filep, "w") as file:
            file.write(code)

        # Reload the Module after its source has been edited
        if self.node_mod is not None:
            reload(self.node_mod)



# mypy: disable-error-code="attr-defined"
import pytest
from frappy.client import CacheItem
from frappy.datatypes import ArrayOf, FloatRange, StringType, StructOf, TupleOf

from secop_ophyd.util import SECoPdtype, SECoPReading

RAGGED = True
REGULAR = False


@pytest.mark.parametrize(
    "start_dtype,data,expected_shape,expected_update_shape",
    [
        pytest.param(
            ArrayOf(FloatRange(), minlen=0, maxlen=100),
            [0.1, 0.1],
            [100],
            [2],
            id="Ragged 1D",
        ),
        pytest.param(
            ArrayOf(ArrayOf(FloatRange()), minlen=0, maxlen=100),
            [[10, 10], [10, 10]],
            [100, 100],  # maxlen defaults to 100 if not given
            [2, 2],
            id="Ragged 2D",
        ),
        pytest.param(
            ArrayOf(ArrayOf(ArrayOf(FloatRange())), minlen=0, maxlen=99),
            [[[10, 10], [10, 10]], [[10, 10], [10, 10]]],
            [99, 100, 100],
            [2, 2, 2],
            id="Ragged 3D",
        ),
        pytest.param(
            ArrayOf(
                ArrayOf(ArrayOf(FloatRange(), minlen=2, maxlen=2), minlen=2, maxlen=2),
                minlen=2,
                maxlen=2,
            ),
            [[[10, 10], [10, 10]], [[10, 10], [10, 10]]],
            [2, 2, 2],
            [2, 2, 2],
            id="Regular 3D",
        ),
        pytest.param(
            ArrayOf(TupleOf(ArrayOf(FloatRange(), minlen=0, maxlen=50), StringType())),
            [([1, 2, 3], "blah")],
            [100],
            [1],
            id="Complex Ragged 1D",
        ),
        pytest.param(
            ArrayOf(
                ArrayOf(
                    TupleOf(ArrayOf(FloatRange(), minlen=0, maxlen=50), StringType())
                )
            ),
            [
                [([1, 2, 3], "blah"), ([1, 2, 3], "blah")],
                [([1, 2, 3], "blah"), ([1, 2, 3], "blah")],
            ],
            [100, 100],
            [2, 2],
            id="Complex Ragged 2D",
        ),
        pytest.param(
            TupleOf(ArrayOf(FloatRange(), minlen=0, maxlen=50), StringType()),
            ([1, 2, 3], "blah"),
            [],
            [],
            id="Complex",
        ),
        pytest.param(
            StructOf(
                mass=ArrayOf(FloatRange()),
                pressure=ArrayOf(FloatRange()),
                timestamp=ArrayOf(FloatRange()),
            ),
            {"mass": [], "pressure": [], "timestamp": []},
            [],
            [],
            id="struct of empty arrays",
        ),
        pytest.param(
            ArrayOf(ArrayOf(FloatRange())),
            [[], []],
            [100, 100],
            [2, 0],
            id="empty 2d array ",
        ),
        pytest.param(
            ArrayOf(ArrayOf(FloatRange())),
            [[]],
            [100, 100],
            [1, 0],
            id="empty 2d array [1,0]",
        ),
        pytest.param(
            ArrayOf(ArrayOf(FloatRange())),
            [],
            [100, 100],
            [0],
            id="empty empty 2d",
        ),
    ],
)
def test_arrayof_update_dtype(start_dtype, data, expected_shape, expected_update_shape):
    sdtype = SECoPdtype(start_dtype)
    assert sdtype.shape == expected_shape

    sdtype.update_dtype(data)
    assert sdtype.shape == expected_update_shape

    SECoPReading(sdtype, CacheItem(data, 234342.2, None, datatype=start_dtype))


@pytest.mark.parametrize(
    "start_dtype,expected_dtype_descr,expected_shape,max_depth",
    [
        pytest.param(
            TupleOf(ArrayOf(FloatRange(), maxlen=50), StringType()),
            [("f0", "<f8", (50,)), ("f1", "<U100")],
            [],
            1,
            id="Tuple of Array",
        ),
        pytest.param(
            ArrayOf(
                TupleOf(FloatRange(), StringType(), ArrayOf(ArrayOf(FloatRange()))),
                maxlen=50,
            ),
            [("f0", "<f8"), ("f1", "<U100"), ("f2", "<f8", (100, 100))],
            [50],
            1,
            id="Array of Tuple",
        ),
        pytest.param(
            ArrayOf(
                TupleOf(
                    FloatRange(), StringType(), TupleOf(FloatRange(), StringType())
                ),
                maxlen=50,
            ),
            [("f0", "<f8"), ("f1", "<U100"), ("f2", [("f0", "<f8"), ("f1", "<U100")])],
            [50],
            2,
            id="Array of (Tuple of Tuple)",
        ),
        pytest.param(
            ArrayOf(
                StructOf(a=FloatRange(), b=StringType(), c=StringType()),
                maxlen=50,
            ),
            [("a", "<f8"), ("b", "<U100"), ("c", "<U100")],
            [50],
            1,
            id="Array of Struct",
        ),
        pytest.param(
            ArrayOf(
                StructOf(
                    a=FloatRange(),
                    b=StringType(),
                    c=TupleOf(FloatRange(), FloatRange()),
                ),
                maxlen=50,
            ),
            [("a", "<f8"), ("b", "<U100"), ("c", [("f0", "<f8"), (("f1", "<f8"))])],
            [50],
            2,
            id="Array of (Struct of tuple)",
        ),
    ],
)
def test_describe_str(start_dtype, expected_dtype_descr, expected_shape, max_depth):
    sdtype = SECoPdtype(start_dtype)
    assert sdtype.shape == expected_shape
    assert sdtype.dtype_descr == expected_dtype_descr
    assert sdtype.max_depth == max_depth



# mypy: disable-error-code="attr-defined"
import warnings

from bluesky.plans import count

from secop_ophyd.SECoPDevices import SECoPNodeDevice

warnings.filterwarnings(
    "ignore",
    message="""The method Broker.insert may be removed
    in a future release of databroker.""",
)


async def test_run_engine_count(cryo_sim, run_engine, cryo_node: SECoPNodeDevice, db):
    p = run_engine(count([cryo_node.cryo], num=5, delay=1))

    run = db[p]

    data = run.primary.read()

    cryo_dat = data[cryo_node.cryo.value.name]

    assert len(cryo_dat) == 5
    assert max(cryo_dat.data) < 11 and min(cryo_dat.data) > 8

    cryo_node.disconnect()


async def test_run_engine_string_value(
    nested_struct_sim, run_engine, nested_node_re: SECoPNodeDevice, db
):
    # p = RE(read(nested_node_RE.str_test))
    nested_node_re.disconnect()



# mypy: disable-error-code="attr-defined"

import pytest
from bluesky import RunEngine
from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky.plans import count
from bluesky.utils import ProgressBarManager
from ophyd_async.core import SignalR

from secop_ophyd.SECoPDevices import SECoPNodeDevice, SECoPReadableDevice


async def test_primitive_arrays(
    nested_struct_sim, run_engine: RunEngine, nested_node_re: SECoPNodeDevice
):
    bec = BestEffortCallback()
    run_engine.subscribe(bec)
    run_engine.waiting_hook = ProgressBarManager()
    run_engine.ignore_callback_exceptions = False

    prim_arr: SECoPReadableDevice = nested_node_re.primitive_arrays

    prim_arr.add_readables(
        [
            prim_arr.arr_int,
            prim_arr.arr_float,
            prim_arr.arr_arr_float,
            prim_arr.arr_bool,
            prim_arr.arr_String,
            prim_arr.arr_String_nomax,
        ]
    )

    run_engine(count([prim_arr], 1, 3))

    nested_node_re.disconnect()


@pytest.mark.parametrize(
    "sig,shape_expected,dtype_numpy_expected",
    [
        pytest.param("arr_int", [6], "<i8", id="integer array"),
        pytest.param("arr_float", [2], "<f8", id="float array"),
        pytest.param("arr_arr_float", [2, 2], "<f8", id="2d float array"),
        pytest.param("arr_bool", [3], "<b1", id="boolean array"),
        pytest.param("arr_String", [3], "<U20", id="string array"),
        pytest.param("arr_String_nomax", [3], "<U100", id="string array no maxlen"),
    ],
)
async def test_primitive_float_array(
    nested_struct_sim,
    run_engine: RunEngine,
    nested_node_re: SECoPNodeDevice,
    shape_expected,
    sig,
    dtype_numpy_expected,
):
    bec = BestEffortCallback()
    run_engine.subscribe(bec)
    run_engine.waiting_hook = ProgressBarManager()
    run_engine.ignore_callback_exceptions = False

    prim_arr: SECoPReadableDevice = nested_node_re.primitive_arrays

    arr_signal: SignalR = getattr(prim_arr, sig)

    descr = await arr_signal.describe()

    decr_dict = descr[arr_signal.name]

    dtype_numpy = decr_dict["dtype_numpy"]
    shape = decr_dict["shape"]

    assert shape == shape_expected
    assert dtype_numpy == dtype_numpy_expected

    nested_node_re.disconnect()
